#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "asm_val.h"
#include "asm_field.h"
#include "asm_entry.h"
#include "asm_list.h"
#include "assemble.h"
#include "lib.h"

static int assemble_arp(pkt_asm_list_t list, char *buffer, int size)
{
  struct arphdr arphdr;
  char *p, *smac, *tmac, *sip, *tip;
  struct arpdata {
    union {
      char *octet;
      struct ether_addr *addr;
    } sender_macaddr;
    union {
      char *octet;
      struct ether_addr *addr;
    } target_macaddr;
    struct in_addr sender_ipaddr;
    struct in_addr target_ipaddr;
  } arpdata;
  pkt_asm_val_t val;

  if (size < sizeof(arphdr))
    return -1;
  memcpy(&arphdr, buffer, sizeof(arphdr));

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_HARDWARE))) {
    arphdr.ar_hrd = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_PROTOCOL))) {
    arphdr.ar_pro = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_HARDWARE_SIZE))) {
    arphdr.ar_hln = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_PROTOCOL_SIZE))) {
    arphdr.ar_pln = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_OPERATION))) {
    arphdr.ar_op = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  if ((ntohs(arphdr.ar_hrd) != ARPHRD_ETHER) ||
      (ntohs(arphdr.ar_pro) != ETHERTYPE_IP) ||
      (arphdr.ar_hln != ETHER_ADDR_LEN) ||
      (arphdr.ar_pln != sizeof(in_addr_t)))
    return -1;

  if (size < sizeof(struct arphdr) + arphdr.ar_hln * 2 + arphdr.ar_pln * 2)
    return -1;

  p = buffer;
  p += sizeof(struct arphdr); smac = p;
  p += arphdr.ar_hln; sip  = p;
  p += arphdr.ar_pln; tmac = p;
  p += arphdr.ar_hln; tip  = p;

  arpdata.sender_macaddr.octet = smac;
  arpdata.target_macaddr.octet = tmac;
  memcpy(&arpdata.sender_ipaddr.s_addr, sip, sizeof(in_addr_t));
  memcpy(&arpdata.target_ipaddr.s_addr, tip, sizeof(in_addr_t));

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_SENDER_MACADDR))) {
    memcpy(arpdata.sender_macaddr.octet,
	   pkt_asm_val_get_value_macaddr(val), ETHER_ADDR_LEN);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_SENDER_IPADDR))) {
    arpdata.sender_ipaddr.s_addr =
      htonl(pkt_asm_val_get_value_ipaddr(val)->s_addr);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_TARGET_MACADDR))) {
    memcpy(arpdata.target_macaddr.octet,
	   pkt_asm_val_get_value_macaddr(val), ETHER_ADDR_LEN);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ARP_TARGET_IPADDR))) {
    arpdata.target_ipaddr.s_addr =
      htonl(pkt_asm_val_get_value_ipaddr(val)->s_addr);
    pkt_asm_val_destroy(val);
  }

  memcpy(sip, &arpdata.sender_ipaddr.s_addr, sizeof(in_addr_t));
  memcpy(tip, &arpdata.target_ipaddr.s_addr, sizeof(in_addr_t));

  memcpy(buffer, &arphdr, sizeof(arphdr));

  return size;
}

static int assemble_icmp(pkt_asm_list_t list, char *buffer, int size, int total_size)
{
  struct icmp *icmphdr;
  pkt_asm_val_t val;

  if ((size < total_size) || (total_size < sizeof(*icmphdr)))
    return -1;
  icmphdr = (struct icmp *)buffer;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMP_TYPE))) {
    icmphdr->icmp_type = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMP_CODE))) {
    icmphdr->icmp_code = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMP_CHECKSUM))) {
    icmphdr->icmp_cksum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  return size;
}

static int assemble_igmp(pkt_asm_list_t list, char *buffer, int size, int total_size)
{
  struct igmp *igmphdr;
  pkt_asm_val_t val;

  if ((size < total_size) || (total_size < sizeof(*igmphdr)))
    return -1;
  if (total_size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IGMP_TYPE))) {
    igmphdr->igmp_type = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IGMP_CODE))) {
    igmphdr->igmp_code = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IGMP_CHECKSUM))) {
    igmphdr->igmp_cksum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IGMP_GROUP))) {
    igmphdr->igmp_group.s_addr =
      htonl(pkt_asm_val_get_value_ipaddr(val)->s_addr);
    pkt_asm_val_destroy(val);
  }

  return size;
}

static int assemble_tcp(pkt_asm_list_t list, char *buffer, int size, int total_size)
{
  struct tcphdr *tcphdr;
  pkt_asm_val_t val;

  if ((size < total_size) || (total_size < sizeof(*tcphdr)))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_SRC_PORT))) {
    tcphdr->th_sport = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_DST_PORT))) {
    tcphdr->th_dport = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_SEQ_NUMBER))) {
    tcphdr->th_seq = htonl(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_ACK_NUMBER))) {
    tcphdr->th_ack = htonl(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_OFFSET))) {
    tcphdr->th_off = pkt_asm_val_get_value_integer(val) >> 2;
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_FLAGS))) {
    tcphdr->th_flags = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_WINDOW))) {
    tcphdr->th_win = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_CHECKSUM))) {
    tcphdr->th_sum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_TCP_URGENT))) {
    tcphdr->th_urp = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  return size;
}

static int assemble_udp(pkt_asm_list_t list, char *buffer, int size, int total_size)
{
  struct udphdr *udphdr;
  pkt_asm_val_t val;

  if ((size < total_size) || (total_size < sizeof(*udphdr)))
    return -1;
  udphdr = (struct udphdr *)buffer;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_UDP_SRC_PORT))) {
    udphdr->uh_sport = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_UDP_DST_PORT))) {
    udphdr->uh_dport = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_UDP_SIZE))) {
    udphdr->uh_ulen = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_UDP_CHECKSUM))) {
    udphdr->uh_sum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  return size;
}

static int assemble_ip(pkt_asm_list_t list, char *buffer, int size)
{
  static char *pktbuf = NULL;
  static int bufsize = 0;
  char *p;
  int s;
  struct ip *iphdr;
  int hdrsize, paysize;
  pkt_asm_val_t val;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (hdrsize < sizeof(*iphdr))
    return -1;
  if (paysize < 0)
    return -1;
  if (size < hdrsize + paysize)
    return -1;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_VERSION))) {
    iphdr->ip_v = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_HEADER_SIZE))) {
    hdrsize = pkt_asm_val_get_value_integer(val);
    iphdr->ip_hl = hdrsize >> 2;
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_TOS))) {
    iphdr->ip_tos = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_TOTAL_SIZE))) {
    iphdr->ip_len = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_ID))) {
    iphdr->ip_id = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_FRAGMENT))) {
    iphdr->ip_off = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_TTL))) {
    iphdr->ip_ttl = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_PROTOCOL))) {
    iphdr->ip_p = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_CHECKSUM))) {
    iphdr->ip_sum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_SRC_ADDR))) {
    iphdr->ip_src.s_addr = htonl(pkt_asm_val_get_value_ipaddr(val)->s_addr);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IP_DST_ADDR))) {
    iphdr->ip_dst.s_addr = htonl(pkt_asm_val_get_value_ipaddr(val)->s_addr);
    pkt_asm_val_destroy(val);
  }

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP: s = assemble_icmp(list, p, s, paysize); break;
  case IPPROTO_IGMP: s = assemble_igmp(list, p, s, paysize); break;
  case IPPROTO_TCP:  s = assemble_tcp( list, p, s, paysize); break;
  case IPPROTO_UDP:  s = assemble_udp( list, p, s, paysize); break;
  default: break;
  }

  if (s < 1)
    return s;

  size = s + (p - pktbuf);

  memcpy(buffer, pktbuf, size);

  return size;
}

static int assemble_icmp6(pkt_asm_list_t list, char *buffer, int size, int total_size)
{
  struct icmp6_hdr *icmp6hdr;
  pkt_asm_val_t val;

  if ((size < total_size) || (total_size < sizeof(*icmp6hdr)))
    return -1;
  icmp6hdr = (struct icmp6_hdr *)buffer;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMPV6_TYPE))) {
    icmp6hdr->icmp6_type = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMPV6_CODE))) {
    icmp6hdr->icmp6_code = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ICMPV6_CHECKSUM))) {
    icmp6hdr->icmp6_cksum = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  return size;
}

static int assemble_ip6(pkt_asm_list_t list, char *buffer, int size)
{
  static char *pktbuf = NULL;
  static int bufsize = 0;
  char *p;
  int s;
  struct ip6_hdr *ip6hdr;
  int hdrsize, paysize;
  int nexthdr, exthdrsize;
  struct ip6_ext *ip6exthdr;
  struct ip6_rthdr *ip6rthdr;
  struct ip6_frag *ip6fraghdr;
  pkt_asm_val_t val;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*ip6hdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  ip6hdr = (struct ip6_hdr *)pktbuf;

  hdrsize = sizeof(struct ip6_hdr);
  paysize = ntohs(ip6hdr->ip6_plen);

  if (size < hdrsize + paysize)
    return -1;

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_VERSION))) {
    ip6hdr->ip6_vfc &= ~0xf0;
    ip6hdr->ip6_vfc |= pkt_asm_val_get_value_integer(val) << 4;
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_FLOWID))) {
    ip6hdr->ip6_flow = htonl(ntohl(ip6hdr->ip6_flow) & ~0xffffff);
    ip6hdr->ip6_flow |= htonl(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_PAYLOAD_SIZE))) {
    paysize = pkt_asm_val_get_value_integer(val);
    ip6hdr->ip6_plen = htons(paysize);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_HOP_LIMIT))) {
    ip6hdr->ip6_hlim = pkt_asm_val_get_value_integer(val);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_SRC_ADDR))) {
    memcpy(&ip6hdr->ip6_src, pkt_asm_val_get_value_ipv6addr(val),
	   sizeof(struct in6_addr));
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_DST_ADDR))) {
    memcpy(&ip6hdr->ip6_dst, pkt_asm_val_get_value_ipv6addr(val),
	   sizeof(struct in6_addr));
    pkt_asm_val_destroy(val);
  }

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  nexthdr = ip6hdr->ip6_nxt;

  while (1) {
    val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_NEXT_HEADER);
    if (val) {
      ip6hdr->ip6_nxt = pkt_asm_val_get_value_integer(val);
      pkt_asm_val_destroy(val);
    }
    exthdrsize = 0;
    switch (nexthdr) {
    case IPPROTO_HOPOPTS:
    case IPPROTO_DSTOPTS:
      if (paysize < sizeof(*ip6exthdr))
	return -1;
      ip6exthdr = (struct ip6_ext *)p;
      nexthdr = ip6exthdr->ip6e_nxt;
      exthdrsize = (ip6exthdr->ip6e_len + 1) << 3;
      break;
    case IPPROTO_ROUTING:
      if (paysize < sizeof(*ip6rthdr))
	return -1;
      ip6rthdr = (struct ip6_rthdr *)p;
      nexthdr = ip6rthdr->ip6r_nxt;
      exthdrsize = (ip6rthdr->ip6r_len + 1) << 3;
      break;
    case IPPROTO_FRAGMENT:
      if (paysize < sizeof(*ip6fraghdr))
	return -1;
      ip6fraghdr = (struct ip6_frag *)p;
      nexthdr = ip6fraghdr->ip6f_nxt;
      exthdrsize = sizeof(*ip6fraghdr);
      break;
    case IPPROTO_NONE:
    default:
      break;
    }
    if (exthdrsize == 0)
      break;
    if (paysize < exthdrsize)
      return -1;
    val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_IPV6_NEXT_SIZE);
    if (val) {
      /* TODO */
      pkt_asm_val_destroy(val);
    }
    p       += exthdrsize;
    s       -= exthdrsize;
    paysize -= exthdrsize;
  }

  switch (nexthdr) {
  case IPPROTO_ICMPV6: s = assemble_icmp6(list, p, s, paysize); break;
  case IPPROTO_TCP:    s = assemble_tcp(  list, p, s, paysize); break;
  case IPPROTO_UDP:    s = assemble_udp(  list, p, s, paysize); break;
  default: break;
  }

  if (s < 1)
    return s;

  size = s + (p - pktbuf);

  memcpy(buffer, pktbuf, size);

  return size;
}

int pkt_assemble_ethernet(pkt_asm_list_t list, char *buffer, int size)
{
  char *p;
  int s;
  struct ether_header ehdr;
  struct {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;
  pkt_asm_val_t val;

  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);

  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ETHERNET_DST_ADDR))) {
    memcpy(ehdr.ether_dhost, pkt_asm_val_get_value_macaddr(val),
	   ETHER_ADDR_LEN);
    pkt_asm_val_destroy(val);
  }
  while ((val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ETHERNET_SRC_ADDR))) {
    memcpy(ehdr.ether_shost, pkt_asm_val_get_value_macaddr(val),
	   ETHER_ADDR_LEN);
    pkt_asm_val_destroy(val);
  }
  val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ETHERNET_TYPE);
  if (val) {
    ehdr.ether_type = htons(pkt_asm_val_get_value_integer(val));
    pkt_asm_val_destroy(val);
  }

  type = ntohs(ehdr.ether_type);
  p = buffer + ETHER_HDR_LEN;
  s = size   - ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86DD
#endif

  while (type == ETHERTYPE_VLAN) {
    if (s < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, p, sizeof(vlantag));
    val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ETHERNET_VLAN_TAG);
    if (val) {
      vlantag.tag = htons(pkt_asm_val_get_value_integer(val));
      pkt_asm_val_destroy(val);
    }
    val = pkt_asm_list_del_val(list, PKT_ASM_FIELD_ETHERNET_TYPE);
    if (val) {
      vlantag.proto = htons(pkt_asm_val_get_value_integer(val));
      pkt_asm_val_destroy(val);
    }
    memcpy(p, &vlantag, sizeof(vlantag));
    type = ntohs(vlantag.proto);
    p += sizeof(vlantag);
    s -= sizeof(vlantag);
  }

  ehdr.ether_type = htons(type);

  switch (type) {
  case ETHERTYPE_ARP:  s = assemble_arp(list, p, s); break;
  case ETHERTYPE_IP:   s = assemble_ip( list, p, s); break;
  case ETHERTYPE_IPV6: s = assemble_ip6(list, p, s); break;
  default: break;
  }

  if (s < 1)
    return s;

  size = s + (p - buffer);

  if (p > buffer + ETHER_HDR_LEN) {
    memcpy(&vlantag, p - sizeof(vlantag), sizeof(vlantag));
    vlantag.proto = ehdr.ether_type;
    memcpy(p - sizeof(vlantag), &vlantag, sizeof(vlantag));
    ehdr.ether_type = htons(ETHERTYPE_VLAN);
  }

  memcpy(buffer, &ehdr, ETHER_HDR_LEN);

  return size;
}
